#ifndef CONTEXT_HPP
#define CONTEXT_HPP

#include "symbol.hpp"
#include "function.hpp"
#include "exceptions.hpp"
#include "numeric_traits.hpp"
#include "cli.hpp"
#include <functional>
#include <vector>
#include <deque>
#include <set>

template<typename ...>
struct register_function_helper;

class cycle_detector;

class interpreter_context {
public :
    interpreter_context(args_t& args);

    typedef std::function<symbol_impl*(const std::vector<symbol>&)>   function_t;
    typedef std::function<symbol_impl*(const symbol&, const symbol&)> binary_operator_t;
    typedef std::function<symbol_impl*(const symbol&)>                unary_operator_t;

    std::string prompt;
    std::string align;
    bool ask_exit;
    int  retcode;

    bool allow_undefined;
    bool allow_complex;
    bool high_precision;
    std::size_t digits;
    std::size_t max_cycles;

    mutable bool evaluate_mode;

private :
    mutable std::deque<std::string> msg_stack;

    std::map<std::string, symbol> vars;

    std::map<std::string, std::map<signature_t, function_t>> funcs;

    std::map<std::string, std::map<type_t, std::map<type_t, binary_operator_t>>> bin_ops;
    std::map<std::string, std::map<type_t, unary_operator_t>>                    un_ops;

    std::vector<std::vector<std::string>> bin_ops_priority;

public :
    void msg(const std::string& s) const;
    void error(const std::string& s) const;
    void error(const no_function_match_exception& e) const;
    void warning(const std::string& s) const;
    void note(const std::string& s) const;
    std::string flush_msg() const;
    void print_help() const;

    void format_for_output(std::string& s) const;

    template<typename ntype>
    std::size_t get_index(const ntype& ti, std::size_t n) const {
        if (ti >= numeric_traits<ntype>::zero) {
            auto i = numeric_traits<ntype>::to_uint(ti);
            if (i < n) return i;
        } else {
            auto i = ti + ntype(n);
            if (i >= numeric_traits<ntype>::zero) {
                return numeric_traits<ntype>::to_uint(i);
            }
        }

        error("index out of range "+numeric_traits<ntype>::to_string(ti, *this)+
            " (size: "+numeric_traits<ntype>::to_string(n, *this)+")");

        throw parser_exception();
    }

    symbol_impl* evaluate_identifier(const std::string& name) const;
    std::string serialize_variable(const std::string& name) const;
    void set_variable(const std::string& name, symbol&& s);
    void erase_variable(const std::string& name);
    void erase_variables();

    template<typename ... Args, bool enable = check_variadic<Args...>::value>
    void register_function(const char* name, const function_t& func) {
        funcs[name][function::make_signature<Args...>()] = func;
    }

    template<typename ... Args, bool enable = check_variadic<Args...>::value>
    void register_function(const std::vector<std::string>& names, const function_t& func) {
        for (const auto& name : names)
            funcs[name][function::make_signature<Args...>()] = func;
    }

    template<typename T1, typename T2>
    void register_operator(const std::string& op, const binary_operator_t& func) {
        bin_ops[op][T1::static_type_id()][T2::static_type_id()] = func;
    }

    template<typename T1, typename T2>
    void register_commutative_operator(const std::string& op, const binary_operator_t& func) {
        bin_ops[op][T1::static_type_id()][T2::static_type_id()] = func;
        bin_ops[op][T2::static_type_id()][T1::static_type_id()] = std::bind(func, std::placeholders::_2, std::placeholders::_1);
    }

    template<typename T>
    void register_operator(const std::string& op, const unary_operator_t& func) {
        un_ops[op][T::static_type_id()] = func;
    }

    symbol_impl* find_signature(std::map<std::string, std::map<signature_t, function_t>>::const_iterator iter_name, const signature_t& sig, const std::vector<symbol>& args) const;
    symbol_impl* call_function(const std::string& name, const std::vector<symbol>& args) const;
    symbol_impl* call_function_no_eval(const std::string& name, const std::vector<symbol>& args) const;
    symbol_impl* binary_operator(const std::string& op, const symbol& ts1, const symbol& ts2) const;
    symbol_impl* unary_operator(const std::string& op, const symbol& ts) const;

    std::pair<std::size_t,std::size_t> binary_operator_priority(const std::string& op) const;
    std::pair<std::size_t,std::size_t> unary_operator_priority(const std::string& op) const;

    symbol_impl* read_complex(const std::string& s) const;
    symbol_impl* read_scalar(const std::string& s) const;

    cycle_detector detect_cycles() const;

private :
    friend class cycle_detector;

    mutable bool                               cycle_detection;
    mutable std::map<std::string, std::size_t> cycle_eval_list;
    void end_cycle_detection() const;

public :
    symbol_impl* evaluate(const std::string&) const;

private :
    friend class math_parser_impl;
    std::function<symbol_impl*(const std::string&)> evaluate_function;
};

class cycle_detector {
    const interpreter_context& context;
    bool                       detect;

public :
    cycle_detector(const interpreter_context& context, bool detect);
    ~cycle_detector();
};

#endif
